r"""Contain a ``polars.DataFrame`` transformer to compute difference
between columns."""

from __future__ import annotations

__all__ = ["AbsDiffHorizontalTransformer", "DiffHorizontalTransformer"]

import logging

import polars as pl

from grizz.transformer.columns import BaseIn2Out1Transformer

logger = logging.getLogger(__name__)


class AbsDiffHorizontalTransformer(BaseIn2Out1Transformer):
    r"""Implement a transformer to compute the absolute difference
    between two columns.

    Internally, this tranformer computes: ``out = abs(in1 - in2)``

    Args:
        in1_col: The first input column name.
        in2_col: The second input column name.
        out_col: The output column name.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import AbsDiffHorizontal
    >>> transformer = AbsDiffHorizontal(in1_col="col1", in2_col="col2", out_col="diff")
    >>> transformer
    AbsDiffHorizontalTransformer(in1_col='col1', in2_col='col2', out_col='diff', exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": [5, 4, 3, 2, 1],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ 5    ┆ a    │
    │ 2    ┆ 4    ┆ b    │
    │ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 2    ┆ d    │
    │ 5    ┆ 1    ┆ e    │
    └──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ diff │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ str  ┆ i64  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 5    ┆ a    ┆ 4    │
    │ 2    ┆ 4    ┆ b    ┆ 2    │
    │ 3    ┆ 3    ┆ c    ┆ 0    │
    │ 4    ┆ 2    ┆ d    ┆ 2    │
    │ 5    ┆ 1    ┆ e    ┆ 4    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Computing the absolute difference between {self._in1_col!r} and "
            f"{self._in2_col!r} | out_col={self._out_col!r}"
        )
        return frame.with_columns(
            frame.select((pl.col(self._in1_col) - pl.col(self._in2_col)).abs().alias(self._out_col))
        )


class DiffHorizontalTransformer(BaseIn2Out1Transformer):
    r"""Implement a transformer to compute the difference between two
    columns.

    Internally, this tranformer computes: ``out = in1 - in2``

    Args:
        in1_col: The first input column name.
        in2_col: The second input column name.
        out_col: The output column name.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import DiffHorizontal
    >>> transformer = DiffHorizontal(in1_col="col1", in2_col="col2", out_col="diff")
    >>> transformer
    DiffHorizontalTransformer(in1_col='col1', in2_col='col2', out_col='diff', exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": [5, 4, 3, 2, 1],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ 5    ┆ a    │
    │ 2    ┆ 4    ┆ b    │
    │ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 2    ┆ d    │
    │ 5    ┆ 1    ┆ e    │
    └──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ diff │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ i64  ┆ str  ┆ i64  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 5    ┆ a    ┆ -4   │
    │ 2    ┆ 4    ┆ b    ┆ -2   │
    │ 3    ┆ 3    ┆ c    ┆ 0    │
    │ 4    ┆ 2    ┆ d    ┆ 2    │
    │ 5    ┆ 1    ┆ e    ┆ 4    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Computing the difference between {self._in1_col!r} and "
            f"{self._in2_col!r} | out_col={self._out_col!r}"
        )
        return frame.with_columns(
            frame.select((pl.col(self._in1_col) - pl.col(self._in2_col)).alias(self._out_col))
        )



r"""Contain ``polars.DataFrame`` transformers to compare element-wise a
DataFrame."""

from __future__ import annotations

__all__ = [
    "BaseComparatorTransformer",
    "EqualMissingTransformer",
    "EqualTransformer",
    "GreaterEqualTransformer",
    "GreaterTransformer",
    "LowerEqualTransformer",
    "LowerTransformer",
    "NotEqualMissingTransformer",
    "NotEqualTransformer",
]

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING, Any

import polars as pl
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseInNTransformer
from grizz.utils.column import check_column_exist_policy, check_existing_columns

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class BaseComparatorTransformer(BaseInNTransformer):
    r"""Define a base class to compare element-wise a DataFrame.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
    """

    def __init__(
        self,
        columns: Sequence[str] | None,
        target: Any,
        prefix: str,
        suffix: str,
        exclude_columns: Sequence[str] = (),
        exist_policy: str = "raise",
        missing_policy: str = "raise",
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._target = target
        self._prefix = prefix
        self._suffix = suffix

        check_column_exist_policy(exist_policy)
        self._exist_policy = exist_policy

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "columns": self._columns,
                "target": self._target,
                "prefix": self._prefix,
                "suffix": self._suffix,
                "exclude_columns": self._exclude_columns,
                "exist_policy": self._exist_policy,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args})"

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self._check_output_columns(frame)
        logger.info(
            f"Applying the {self._get_operation_name()} operation on "
            f"{len(self.find_columns(frame)):,} columns | "
            f"prefix={self._prefix!r} | suffix={self._suffix!r}"
        )
        columns = self.find_common_columns(frame)
        data = self._compare(frame.select(columns))
        return frame.with_columns(data.rename(lambda col: f"{self._prefix}{col}{self._suffix}"))

    def _check_output_columns(self, frame: pl.DataFrame) -> None:
        r"""Check if the output columns already exist.

        Args:
            frame: The input DataFrame to check.
        """
        check_existing_columns(
            frame,
            columns=[f"{self._prefix}{col}{self._suffix}" for col in self.find_columns(frame)],
            exist_policy=self._exist_policy,
        )

    @abstractmethod
    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        r"""Generate the comparison results.

        Args:
            frame: The DataFrame to compare.

        Returns:
            A DataFrame with the same shape and columns as the input,
                but that contains the result of the comparison.
        """

    @abstractmethod
    def _get_operation_name(self) -> str:
        r"""Get the operation name.

        Returns:
            The operation name.
        """


class EqualTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the equal operation.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Equal
    >>> transformer = Equal(columns=["col1", "col3"], target=3, prefix="", suffix="_ind")
    >>> transformer
    EqualTransformer(columns=('col1', 'col3'), target=3, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ false    ┆ false    │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ false    ┆ false    │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ true     ┆ false    │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ false    ┆ false    │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ false    ┆ false    │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().eq(self._target))

    def _get_operation_name(self) -> str:
        return "equal"


class EqualMissingTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the equal operation where
    null values are not propagated.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import EqualMissing
    >>> transformer = EqualMissing(columns=["col1", "col3"], target=3, prefix="", suffix="_ind")
    >>> transformer
    EqualMissingTransformer(columns=('col1', 'col3'), target=3, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ false    ┆ false    │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ false    ┆ false    │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ true     ┆ false    │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ false    ┆ false    │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ false    ┆ false    │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().eq_missing(self._target))

    def _get_operation_name(self) -> str:
        return "equal missing"


class GreaterEqualTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the greater than or equal
    operation.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import GreaterEqual
    >>> transformer = GreaterEqual(
    ...     columns=["col1", "col3"], target=4.2, prefix="", suffix="_ind"
    ... )
    >>> transformer
    GreaterEqualTransformer(columns=('col1', 'col3'), target=4.2, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ false    ┆ true     │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ false    ┆ true     │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ false    ┆ true     │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ false    ┆ true     │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ true     ┆ true     │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().ge(self._target))

    def _get_operation_name(self) -> str:
        return "greater than or equal"


class GreaterTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the greater than
    operation.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Greater
    >>> transformer = Greater(columns=["col1", "col3"], target=4.2, prefix="", suffix="_ind")
    >>> transformer
    GreaterTransformer(columns=('col1', 'col3'), target=4.2, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ false    ┆ true     │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ false    ┆ true     │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ false    ┆ true     │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ false    ┆ true     │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ true     ┆ true     │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().gt(self._target))

    def _get_operation_name(self) -> str:
        return "greater than"


class LowerEqualTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the lower than or equal
    operation.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import LowerEqual
    >>> transformer = LowerEqual(columns=["col1", "col3"], target=4.2, prefix="", suffix="_ind")
    >>> transformer
    LowerEqualTransformer(columns=('col1', 'col3'), target=4.2, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ true     ┆ false    │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ true     ┆ false    │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ true     ┆ false    │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ true     ┆ false    │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ false    ┆ false    │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().le(self._target))

    def _get_operation_name(self) -> str:
        return "lower than or equal"


class LowerTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the lower operation.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Lower
    >>> transformer = Lower(columns=["col1", "col3"], target=4.2, prefix="", suffix="_ind")
    >>> transformer
    LowerTransformer(columns=('col1', 'col3'), target=4.2, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ true     ┆ false    │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ true     ┆ false    │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ true     ┆ false    │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ true     ┆ false    │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ false    ┆ false    │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().lt(self._target))

    def _get_operation_name(self) -> str:
        return "lower than"


class NotEqualTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the not equal operation.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import NotEqual
    >>> transformer = NotEqual(columns=["col1", "col3"], target=3, prefix="", suffix="_ind")
    >>> transformer
    NotEqualTransformer(columns=('col1', 'col3'), target=3, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ true     ┆ true     │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ true     ┆ true     │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ false    ┆ true     │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ true     ┆ true     │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ true     ┆ true     │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().ne(self._target))

    def _get_operation_name(self) -> str:
        return "not equal"


class NotEqualMissingTransformer(BaseComparatorTransformer):
    r"""Implements a transformer that computes the not equal operation
    where where null values are not propagated.

    Args:
        columns: The columns to compare. ``None`` means all the
            columns.
        target: The target value to compare with.
        prefix: The column name prefix for the output columns.
        suffix: The column name suffix for the output columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import NotEqualMissing
    >>> transformer = NotEqualMissing(
    ...     columns=["col1", "col3"], target=3, prefix="", suffix="_ind"
    ... )
    >>> transformer
    NotEqualMissingTransformer(columns=('col1', 'col3'), target=3, prefix='', suffix='_ind', exclude_columns=(), exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": [10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    │
    │ 2    ┆ 2    ┆ 20   ┆ b    │
    │ 3    ┆ 3    ┆ 30   ┆ c    │
    │ 4    ┆ 4    ┆ 40   ┆ d    │
    │ 5    ┆ 5    ┆ 50   ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 6)
    ┌──────┬──────┬──────┬──────┬──────────┬──────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_ind ┆ col3_ind │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---      ┆ ---      │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ bool     ┆ bool     │
    ╞══════╪══════╪══════╪══════╪══════════╪══════════╡
    │ 1    ┆ 1    ┆ 10   ┆ a    ┆ true     ┆ true     │
    │ 2    ┆ 2    ┆ 20   ┆ b    ┆ true     ┆ true     │
    │ 3    ┆ 3    ┆ 30   ┆ c    ┆ false    ┆ true     │
    │ 4    ┆ 4    ┆ 40   ┆ d    ┆ true     ┆ true     │
    │ 5    ┆ 5    ┆ 50   ┆ e    ┆ true     ┆ true     │
    └──────┴──────┴──────┴──────┴──────────┴──────────┘

    ```
    """

    def _compare(self, frame: pl.DataFrame) -> pl.DataFrame:
        return frame.select(pl.all().ne_missing(self._target))

    def _get_operation_name(self) -> str:
        return "not equal missing"



r"""Contain a transformer to combine sequentially multiple
transformers."""

from __future__ import annotations

__all__ = ["SequentialTransformer"]

from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_sequence, str_indent, str_sequence

from grizz.transformer.base import BaseTransformer, setup_transformer

if TYPE_CHECKING:
    from collections.abc import Sequence

    import polars as pl


class SequentialTransformer(BaseTransformer):
    r"""Implement a ``polars.DataFrame`` transformer to apply
    sequentially several transformers.

    Args:
        transformers: The transformers or their configurations.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import (
    ...     Sequential,
    ...     Cast,
    ... )
    >>> transformer = Sequential(
    ...     [
    ...         Cast(columns=["col1"], dtype=pl.Float32),
    ...         Cast(columns=["col2"], dtype=pl.Int64),
    ...     ]
    ... )
    >>> transformer
    SequentialTransformer(
      (0): CastTransformer(columns=('col1',), dtype=Float32, exclude_columns=(), missing_policy='raise')
      (1): CastTransformer(columns=('col2',), dtype=Int64, exclude_columns=(), missing_policy='raise')
    )
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬───────┬───────┐
    │ col1 ┆ col2 ┆ col3  ┆ col4  │
    │ ---  ┆ ---  ┆ ---   ┆ ---   │
    │ i64  ┆ str  ┆ str   ┆ str   │
    ╞══════╪══════╪═══════╪═══════╡
    │ 1    ┆ 1    ┆ a     ┆ a     │
    │ 2    ┆ 2    ┆  b    ┆  b    │
    │ 3    ┆ 3    ┆   c   ┆   c   │
    │ 4    ┆ 4    ┆ d     ┆ d     │
    │ 5    ┆ 5    ┆ e     ┆ e     │
    └──────┴──────┴───────┴───────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬───────┬───────┐
    │ col1 ┆ col2 ┆ col3  ┆ col4  │
    │ ---  ┆ ---  ┆ ---   ┆ ---   │
    │ f32  ┆ i64  ┆ str   ┆ str   │
    ╞══════╪══════╪═══════╪═══════╡
    │ 1.0  ┆ 1    ┆ a     ┆ a     │
    │ 2.0  ┆ 2    ┆  b    ┆  b    │
    │ 3.0  ┆ 3    ┆   c   ┆   c   │
    │ 4.0  ┆ 4    ┆ d     ┆ d     │
    │ 5.0  ┆ 5    ┆ e     ┆ e     │
    └──────┴──────┴───────┴───────┘

    ```
    """

    def __init__(self, transformers: Sequence[BaseTransformer | dict]) -> None:
        self._transformers = tuple(setup_transformer(transformer) for transformer in transformers)

    def __repr__(self) -> str:
        args = ""
        if self._transformers:
            args = f"\n  {repr_indent(repr_sequence(self._transformers))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def __str__(self) -> str:
        args = ""
        if self._transformers:
            args = f"\n  {str_indent(str_sequence(self._transformers))}\n"
        return f"{self.__class__.__qualname__}({args})"

    def fit(self, frame: pl.DataFrame) -> None:
        for transformer in self._transformers:
            transformer.fit(frame)

    def fit_transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        for transformer in self._transformers:
            frame = transformer.fit_transform(frame)
        return frame

    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        for transformer in self._transformers:
            frame = transformer.transform(frame)
        return frame



r"""Contain utility functions to manipulate null values in
DataFrames."""

from __future__ import annotations

__all__ = ["compute_null", "compute_null_count", "compute_temporal_null_count", "propagate_nulls"]


from typing import TYPE_CHECKING
from unittest.mock import Mock

import polars as pl
import polars.selectors as cs
from coola.utils import check_numpy, is_numpy_available

from grizz.utils.temporal import to_step_names

if TYPE_CHECKING:
    from collections.abc import Sequence


if is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    np = Mock()


def compute_null(frame: pl.DataFrame) -> pl.DataFrame:
    r"""Return the number and percentage of null values per column.

    Args:
        frame: The DataFrame to analyze.

    Returns:
        A DataFrame with the number and percentage of null values per
            column.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.null import compute_null
    >>> frame = compute_null(
    ...     pl.DataFrame(
    ...         {
    ...             "int": [None, 1, 0, 1],
    ...             "float": [1.2, 4.2, None, 2.2],
    ...             "str": ["A", "B", None, None],
    ...         },
    ...         schema={"int": pl.Int64, "float": pl.Float64, "str": pl.String},
    ...     )
    ... )
    >>> frame
    shape: (3, 4)
    ┌────────┬──────┬───────┬──────────┐
    │ column ┆ null ┆ total ┆ null_pct │
    │ ---    ┆ ---  ┆ ---   ┆ ---      │
    │ str    ┆ i64  ┆ i64   ┆ f64      │
    ╞════════╪══════╪═══════╪══════════╡
    │ int    ┆ 1    ┆ 4     ┆ 0.25     │
    │ float  ┆ 1    ┆ 4     ┆ 0.25     │
    │ str    ┆ 2    ┆ 4     ┆ 0.5      │
    └────────┴──────┴───────┴──────────┘

    ```
    """
    check_numpy()
    null_count = compute_null_count(frame)
    total_count = np.full((frame.shape[1],), frame.shape[0], dtype=np.int64)
    with np.errstate(invalid="ignore"):
        null_pct = null_count.astype(np.float64) / total_count.astype(np.float64)
    return pl.DataFrame(
        {
            "column": list(frame.columns),
            "null": null_count,
            "total": total_count,
            "null_pct": null_pct,
        },
        schema={"column": pl.String, "null": pl.Int64, "total": pl.Int64, "null_pct": pl.Float64},
    )


def compute_null_count(frame: pl.DataFrame) -> np.ndarray:
    r"""Return the number of null values in each column.

    Args:
        frame: The DataFrame to analyze.

    Returns:
        An array with the number of null values in each column.
            The shape of the array is the number of columns.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.null import compute_null_count
    >>> frame = pl.DataFrame(
    ...     {
    ...         "int": [None, 1, 0, 1],
    ...         "float": [1.2, 4.2, None, 2.2],
    ...         "str": ["A", "B", None, None],
    ...     },
    ...     schema={"int": pl.Int64, "float": pl.Float64, "str": pl.String},
    ... )
    >>> count = compute_null_count(frame)
    >>> count
    array([1, 1, 2])

    ```
    """
    check_numpy()
    if (ncols := frame.shape[1]) == 0:
        return np.zeros(ncols, dtype=np.int64)
    return frame.null_count().to_numpy()[0].astype(int)


def compute_temporal_null_count(
    frame: pl.DataFrame,
    columns: Sequence[str],
    temporal_column: str,
    period: str,
) -> tuple[np.ndarray, np.ndarray, list]:
    r"""Compute the number of null values per temporal segments.

    Args:
        frame: The DataFrame to analyze.
        columns: The list of columns to analyze.
        temporal_column: The temporal column used to analyze
            the temporal distribution.
        period: The temporal period e.g. monthly or daily.

    Returns:
        A tuple with 3 values. The first value is a numpy NDArray
            that contains the number of null values per period. The
            second value is a numpy NDArray that contains the total
            number of values. The third value is a list that contains
            the label of each period.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from grizz.utils.null import compute_temporal_null_count
    >>> nulls, totals, labels = compute_temporal_null_count(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col1": [None, float("nan"), 0.0, 1.0],
    ...             "col2": [None, 1, 0, None],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col1": pl.Float64,
    ...             "col2": pl.Int64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     columns=["col1", "col2"],
    ...     temporal_column="datetime",
    ...     period="1mo",
    ... )
    >>> nulls
    array([2, 0, 0, 1])
    >>> totals
    array([2, 2, 2, 2])
    >>> labels
    ['2020-01', '2020-02', '2020-03', '2020-04']

    ```
    """
    check_numpy()
    frame_na = frame.select(cs.by_name(columns).is_null().cast(pl.Int64), pl.col(temporal_column))
    groups = frame_na.sort(temporal_column).group_by_dynamic(temporal_column, every=period)
    steps = to_step_names(groups=groups, period=period)

    nulls = np.zeros(len(steps), dtype=np.int64)
    totals = np.zeros(len(steps), dtype=np.int64)
    if columns:
        nulls += (
            groups.agg(cs.by_name(columns).sum()).drop(temporal_column).sum_horizontal().to_numpy()
        )
        totals += (
            groups.agg(cs.by_name(columns).count())
            .drop(temporal_column)
            .sum_horizontal()
            .to_numpy()
        )
    return nulls, totals, steps


def propagate_nulls(frame: pl.DataFrame, frame_with_null: pl.DataFrame) -> pl.DataFrame:
    r"""Propagate the null values from ``frame_with_null`` to ``frame``.

    Args:
        frame: The input DataFrame where to add ``None`` values based
            on ``frame_with_null``.
        frame_with_null: The DataFrame with the ``None`` values to
            propagate to ``frame``.

    Returns:
        The output DataFrame.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.null import propagate_nulls
    >>> frame_with_null = pl.DataFrame(
    ...     {
    ...         "col1": [1, None, 3, float("nan"), 5],
    ...         "col2": ["1", "2", None, "4", "5"],
    ...         "col3": [10, 20, 30, None, 50],
    ...     },
    ...     schema={"col1": pl.Float32, "col2": pl.String, "col3": pl.Int64},
    ... )
    >>> frame = frame_with_null.fill_null(99).fill_nan(99)
    >>> frame
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ f32  ┆ str  ┆ i64  │
    ╞══════╪══════╪══════╡
    │ 1.0  ┆ 1    ┆ 10   │
    │ 99.0 ┆ 2    ┆ 20   │
    │ 3.0  ┆ null ┆ 30   │
    │ 99.0 ┆ 4    ┆ 99   │
    │ 5.0  ┆ 5    ┆ 50   │
    └──────┴──────┴──────┘
    >>> out = propagate_nulls(frame=frame, frame_with_null=frame_with_null)
    >>> out
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ f32  ┆ str  ┆ i64  │
    ╞══════╪══════╪══════╡
    │ 1.0  ┆ 1    ┆ 10   │
    │ null ┆ 2    ┆ 20   │
    │ 3.0  ┆ null ┆ 30   │
    │ 99.0 ┆ 4    ┆ null │
    │ 5.0  ┆ 5    ┆ 50   │
    └──────┴──────┴──────┘

    ```
    """
    columns = frame.columns
    return (
        frame.with_columns(frame_with_null.select(pl.all().is_null().name.suffix("__@@isnull@@_")))
        .with_columns(
            pl.when(~pl.col(col + "__@@isnull@@_")).then(pl.col(col)).otherwise(None)
            for col in columns
        )
        .select(columns)
    )



r"""Contain a transformer that executes a SQL query against the
DataFrame."""

from __future__ import annotations

__all__ = ["SqlTransformer"]

import logging
from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from grizz.transformer.base import BaseTransformer
from grizz.utils.format import str_shape_diff

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class SqlTransformer(BaseTransformer):
    r"""Implement a transformer that executes a SQL query against the
    DataFrame.

    Args:
        query: The SQL query to execute.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import SqlTransformer
    >>> transformer = SqlTransformer(query="SELECT col1, col4 FROM self WHERE col1 > 2")
    >>> transformer
    SqlTransformer(
      (query): SELECT col1, col4 FROM self WHERE col1 > 2
    )
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> out = transformer.transform(frame)
    >>> out
    shape: (3, 2)
    ┌──────┬──────┐
    │ col1 ┆ col4 │
    │ ---  ┆ ---  │
    │ i64  ┆ str  │
    ╞══════╪══════╡
    │ 3    ┆ c    │
    │ 4    ┆ d    │
    │ 5    ┆ e    │
    └──────┴──────┘

    ```
    """

    def __init__(self, query: str) -> None:
        self._query = query

    def __repr__(self) -> str:
        args = repr_indent(repr_mapping({"query": self._query}))
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def fit_transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self.fit(frame)
        return self.transform(frame)

    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Executing the following SQL query:\n{self._query}")
        out = frame.sql(self._query)
        logger.info(str_shape_diff(orig=frame.shape, final=out.shape))
        return out



r"""Contain the implementation of a in-memory DataFrame exporter and
ingestor."""

from __future__ import annotations

__all__ = ["InMemoryExporter"]

import logging
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal

from grizz.exceptions import DataFrameNotFoundError
from grizz.exporter.base import BaseExporter
from grizz.ingestor.base import BaseIngestor

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class InMemoryExporter(BaseExporter, BaseIngestor):
    r"""Implement an in-memory DataFrame exporter and ingestor.

    Notes:
        This exporter is both exporter and ingestor as the object
            stores the DataFrame.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.exporter import InMemoryExporter
    >>> exporter = InMemoryExporter()
    >>> exporter
    InMemoryExporter(frame=None)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> exporter.export(frame)
    >>> df = exporter.ingest()
    >>> df
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ str  │
    ╞══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┘

    ```
    """

    def __init__(self) -> None:
        self._frame = None

    def __repr__(self) -> str:
        frame = None if self._frame is None else self._frame.shape
        return f"{self.__class__.__qualname__}(frame={frame})"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._frame, other._frame, equal_nan=equal_nan)

    def export(self, frame: pl.DataFrame) -> None:
        logger.info(f"Exporting the DataFrame (shape={frame.shape}) to memory ...")
        self._frame = frame

    def ingest(self) -> pl.DataFrame:
        logger.info("Ingesting DataFrame...")
        if self._frame is None:
            msg = (
                "No DataFrame available for ingestion. You must export a DataFrame "
                "before to ingest it"
            )
            raise DataFrameNotFoundError(msg)
        return self._frame



r"""Contain a wrapper around an exporter to transform the DataFrame
before to export it."""

from __future__ import annotations

__all__ = ["TransformExporter"]

import logging
from typing import TYPE_CHECKING, Any

from coola.utils import repr_indent, repr_mapping

from grizz.exporter.base import BaseExporter, setup_exporter
from grizz.transformer.base import BaseTransformer, setup_transformer

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class TransformExporter(BaseExporter):
    r"""Implement an exporter that transforms the DataFrame before to
    export it.

    Args:
        transformer: The ``polars.DataFrame`` transformer or
            its configuration.
        exporter: The DataFrame exporter or its configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.exporter import TransformExporter, ParquetExporter
    >>> from grizz.transformer import Cast
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> exporter = TransformExporter(
    ...     transformer=Cast(columns=["col1", "col3"], dtype=pl.Float32),
    ...     exporter=ParquetExporter(path="/path/to/frame.parquet"),
    ... )
    >>> exporter
    TransformExporter(
      (transformer): CastTransformer(columns=('col1', 'col3'), dtype=Float32, exclude_columns=(), missing_policy='raise')
      (exporter): ParquetExporter(path=/path/to/frame.parquet)
    )
    >>> exporter.export(frame)  # doctest: +SKIP


    ```
    """

    def __init__(
        self,
        transformer: BaseTransformer | dict,
        exporter: BaseExporter | dict,
    ) -> None:
        self._transformer = setup_transformer(transformer)
        self._exporter = setup_exporter(exporter)

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping({"transformer": self._transformer, "exporter": self._exporter})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._exporter.equal(other._exporter, equal_nan=equal_nan)

    def export(self, frame: pl.DataFrame) -> None:
        frame = self._transformer.transform(frame)
        self._exporter.export(frame)



r"""Contain ``polars.DataFrame`` transformers to encode the labels in a
given column."""

from __future__ import annotations

__all__ = ["LabelEncoderTransformer"]

import logging

import polars as pl

from grizz.transformer.columns import BaseIn1Out1Transformer
from grizz.utils.imports import check_sklearn, is_sklearn_available

if is_sklearn_available():  # pragma: no cover
    from sklearn.preprocessing import LabelEncoder

logger = logging.getLogger(__name__)


class LabelEncoderTransformer(BaseIn1Out1Transformer):
    r"""Implement a ``polars.DataFrame`` to encode the labels in a given
    column.

    Args:
        in_col: The input column name i.e. the column with the label
            to encode.
        out_col: The output column name i.e. the column with encoded
            labels.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import LabelEncoderTransformer
    >>> transformer = LabelEncoderTransformer(in_col="col1", out_col="out")
    >>> transformer
    LabelEncoderTransformer(in_col='col1', out_col='out', exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": ["a", "b", "c", "d", "e"],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...     }
    ... )
    >>> frame
    shape: (5, 2)
    ┌──────┬──────┐
    │ col1 ┆ col2 │
    │ ---  ┆ ---  │
    │ str  ┆ str  │
    ╞══════╪══════╡
    │ a    ┆ 1    │
    │ b    ┆ 2    │
    │ c    ┆ 3    │
    │ d    ┆ 4    │
    │ e    ┆ 5    │
    └──────┴──────┘
    >>> out = transformer.fit_transform(frame)
    >>> out
    shape: (5, 3)
    ┌──────┬──────┬─────┐
    │ col1 ┆ col2 ┆ out │
    │ ---  ┆ ---  ┆ --- │
    │ str  ┆ str  ┆ i64 │
    ╞══════╪══════╪═════╡
    │ a    ┆ 1    ┆ 0   │
    │ b    ┆ 2    ┆ 1   │
    │ c    ┆ 3    ┆ 2   │
    │ d    ┆ 4    ┆ 3   │
    │ e    ┆ 5    ┆ 4   │
    └──────┴──────┴─────┘

    ```
    """

    def __init__(
        self,
        in_col: str,
        out_col: str,
        exist_policy: str = "raise",
        missing_policy: str = "raise",
    ) -> None:
        super().__init__(
            in_col=in_col,
            out_col=out_col,
            exist_policy=exist_policy,
            missing_policy=missing_policy,
        )

        check_sklearn()
        self._encoder = LabelEncoder()

    def _fit(self, frame: pl.DataFrame) -> None:
        logger.info(f"Fitting the label encoder to the data in column {self._in_col!r}")
        self._encoder.fit(frame[self._in_col].to_numpy())

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(
            f"Encoding labels in {self._in_col!r} and saving output in {self._out_col!r} ..."
        )
        y = self._encoder.transform(frame[self._in_col].to_numpy())
        return frame.with_columns(pl.from_numpy(y, schema=[self._out_col]))



r"""Contain DataFrame exporters."""

from __future__ import annotations

__all__ = [
    "BaseExporter",
    "CsvExporter",
    "InMemoryExporter",
    "ParquetExporter",
    "TransformExporter",
    "is_exporter_config",
    "setup_exporter",
]

from grizz.exporter.base import BaseExporter, is_exporter_config, setup_exporter
from grizz.exporter.csv import CsvExporter
from grizz.exporter.in_memory import InMemoryExporter
from grizz.exporter.parquet import ParquetExporter
from grizz.exporter.transform import TransformExporter



r"""Contain a wrapper around an ingestor to transform the data after
ingestion."""

from __future__ import annotations

__all__ = ["TransformIngestor"]

import logging
from typing import TYPE_CHECKING, Any

from coola.utils import repr_indent, repr_mapping

from grizz.ingestor.base import BaseIngestor, setup_ingestor
from grizz.transformer.base import BaseTransformer, setup_transformer

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class TransformIngestor(BaseIngestor):
    r"""Implement an ingestor that also transforms the DataFrame.

    Args:
        ingestor: The base ingestor.
        transformer: The ``polars.DataFrame`` transformer or
            its configuration.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.ingestor import TransformIngestor, Ingestor
    >>> from grizz.transformer import Cast
    >>> ingestor = TransformIngestor(
    ...     ingestor=Ingestor(
    ...         pl.DataFrame(
    ...             {
    ...                 "col1": ["1", "2", "3", "4", "5"],
    ...                 "col2": ["a", "b", "c", "d", "e"],
    ...                 "col3": [1.2, 2.2, 3.2, 4.2, 5.2],
    ...             }
    ...         )
    ...     ),
    ...     transformer=Cast(columns=["col1", "col3"], dtype=pl.Float32),
    ... )
    >>> ingestor
    TransformIngestor(
      (ingestor): Ingestor(shape=(5, 3))
      (transformer): CastTransformer(columns=('col1', 'col3'), dtype=Float32, exclude_columns=(), missing_policy='raise')
    )
    >>> frame = ingestor.ingest()
    >>> frame
    shape: (5, 3)
    ┌──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 │
    │ ---  ┆ ---  ┆ ---  │
    │ f32  ┆ str  ┆ f32  │
    ╞══════╪══════╪══════╡
    │ 1.0  ┆ a    ┆ 1.2  │
    │ 2.0  ┆ b    ┆ 2.2  │
    │ 3.0  ┆ c    ┆ 3.2  │
    │ 4.0  ┆ d    ┆ 4.2  │
    │ 5.0  ┆ e    ┆ 5.2  │
    └──────┴──────┴──────┘

    ```
    """

    def __init__(self, ingestor: BaseIngestor | dict, transformer: BaseTransformer | dict) -> None:
        self._ingestor = setup_ingestor(ingestor)
        self._transformer = setup_transformer(transformer)

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping({"ingestor": self._ingestor, "transformer": self._transformer})
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._ingestor.equal(other._ingestor, equal_nan=equal_nan)

    def ingest(self) -> pl.DataFrame:
        frame = self._ingestor.ingest()
        return self._transformer.transform(frame)



r"""Contain utility functions to compute the co-occurrence matrix."""

from __future__ import annotations

__all__ = ["compute_pairwise_cooccurrence"]

from unittest.mock import Mock

import polars as pl
from coola.utils import check_numpy, is_numpy_available

if is_numpy_available():
    import numpy as np
else:  # pragma: no cover
    np = Mock()


def compute_pairwise_cooccurrence(frame: pl.DataFrame, ignore_self: bool = False) -> np.ndarray:
    r"""Compute the pairwise column co-occurrence.

    Args:
        frame: The input DataFrame. The column values are expected to
            be 0/1 or true/false.
        ignore_self: If ``True``, the diagonal of the co-occurrence
            matrix (a.k.a. self-co-occurrence) is set to 0.

    Returns:
        The co-occurrence matrix.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.cooccurrence import compute_pairwise_cooccurrence
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 1, 0, 0, 1, 0],
    ...         "col2": [0, 1, 0, 1, 0, 1, 0],
    ...         "col3": [0, 0, 0, 0, 1, 1, 1],
    ...     }
    ... )
    >>> compute_pairwise_cooccurrence(frame)
    array([[3, 2, 1],
           [2, 3, 1],
           [1, 1, 3]])
    >>> compute_pairwise_cooccurrence(frame, ignore_self=True)
    array([[0, 2, 1],
           [2, 0, 1],
           [1, 1, 0]])

    ```
    """
    check_numpy()
    if frame.shape[1] == 0:
        return np.zeros((0, 0), dtype=int)
    data = frame.cast(pl.Boolean).fill_null(False).to_numpy().astype(int)
    co = data.transpose().dot(data)
    if ignore_self:
        np.fill_diagonal(co, 0)
    return co



r"""Contain the implementation of a CSV DataFrame exporter."""

from __future__ import annotations

__all__ = ["CsvExporter"]

import logging
from typing import TYPE_CHECKING, Any

from coola import objects_are_equal

from grizz.exporter.base import BaseExporter
from grizz.utils.format import str_kwargs
from grizz.utils.path import human_file_size, sanitize_path

if TYPE_CHECKING:
    from pathlib import Path

    import polars as pl

logger = logging.getLogger(__name__)


class CsvExporter(BaseExporter):
    r"""Implement a CSV DataFrame exporter.

    Args:
        path: The path to the csv file to ingest.
        **kwargs: Additional keyword arguments for
            ``polars.DataFrame.write_csv``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.exporter import CsvExporter
    >>> exporter = CsvExporter(path="/path/to/frame.csv")
    >>> exporter
    CsvExporter(path=/path/to/frame.csv)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> exporter.export(frame)  # doctest: +SKIP

    ```
    """

    def __init__(self, path: Path | str, **kwargs: Any) -> None:
        self._path = sanitize_path(path)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(path={self._path}{str_kwargs(self._kwargs)})"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._path == other._path and objects_are_equal(
            self._kwargs, other._kwargs, equal_nan=equal_nan
        )

    def export(self, frame: pl.DataFrame) -> None:
        logger.info(f"Exporting the DataFrame of shape={frame.shape} to CSV file {self._path} ...")
        self._path.parent.mkdir(parents=True, exist_ok=True)
        frame.write_csv(self._path, **self._kwargs)
        logger.info(f"DataFrame exported | size={human_file_size(self._path)}")



r"""Contain ``polars.DataFrame`` transformers to scale each column to a
given range."""

from __future__ import annotations

__all__ = ["MinMaxScalerTransformer"]

import logging
from typing import TYPE_CHECKING, Any

import polars as pl
from coola.utils.format import repr_mapping_line

from grizz.transformer.columns import BaseInNTransformer
from grizz.utils.column import check_column_exist_policy, check_existing_columns
from grizz.utils.format import str_kwargs
from grizz.utils.imports import check_sklearn, is_sklearn_available
from grizz.utils.null import propagate_nulls

if is_sklearn_available():  # pragma: no cover
    import sklearn

if TYPE_CHECKING:
    from collections.abc import Sequence

logger = logging.getLogger(__name__)


class MinMaxScalerTransformer(BaseInNTransformer):
    r"""Implement a transformer to scale each column to a given range.

    Args:
        columns: The columns to scale. ``None`` means all the
            columns.
        prefix: The column name prefix for the copied columns.
        suffix: The column name suffix for the copied columns.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        propagate_nulls: If set to ``True``, the ``None`` values are
            propagated after the transformation. If ``False``, the
            ``None`` values are replaced by NaNs.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: Additional arguments passed to
            ``sklearn.preprocessing.MinMaxScaler``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import MinMaxScaler
    >>> transformer = MinMaxScaler(columns=["col1", "col3"], prefix="", suffix="_scaled")
    >>> transformer
    MinMaxScalerTransformer(columns=('col1', 'col3'), prefix='', suffix='_scaled', exclude_columns=(), propagate_nulls=True, exist_policy='raise', missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [0, 1, 2, 3, 4, 5],
    ...         "col2": ["0", "1", "2", "3", "4", "5"],
    ...         "col3": [0, 10, 20, 30, 40, 50],
    ...         "col4": ["a", "b", "c", "d", "e", "f"],
    ...     }
    ... )
    >>> frame
    shape: (6, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ i64  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 0    ┆ 0    ┆ 0    ┆ a    │
    │ 1    ┆ 1    ┆ 10   ┆ b    │
    │ 2    ┆ 2    ┆ 20   ┆ c    │
    │ 3    ┆ 3    ┆ 30   ┆ d    │
    │ 4    ┆ 4    ┆ 40   ┆ e    │
    │ 5    ┆ 5    ┆ 50   ┆ f    │
    └──────┴──────┴──────┴──────┘

    >>> out = transformer.fit_transform(frame)
    >>> out
    shape: (6, 6)
    ┌──────┬──────┬──────┬──────┬─────────────┬─────────────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 ┆ col1_scaled ┆ col3_scaled │
    │ ---  ┆ ---  ┆ ---  ┆ ---  ┆ ---         ┆ ---         │
    │ i64  ┆ str  ┆ i64  ┆ str  ┆ f64         ┆ f64         │
    ╞══════╪══════╪══════╪══════╪═════════════╪═════════════╡
    │ 0    ┆ 0    ┆ 0    ┆ a    ┆ 0.0         ┆ 0.0         │
    │ 1    ┆ 1    ┆ 10   ┆ b    ┆ 0.2         ┆ 0.2         │
    │ 2    ┆ 2    ┆ 20   ┆ c    ┆ 0.4         ┆ 0.4         │
    │ 3    ┆ 3    ┆ 30   ┆ d    ┆ 0.6         ┆ 0.6         │
    │ 4    ┆ 4    ┆ 40   ┆ e    ┆ 0.8         ┆ 0.8         │
    │ 5    ┆ 5    ┆ 50   ┆ f    ┆ 1.0         ┆ 1.0         │
    └──────┴──────┴──────┴──────┴─────────────┴─────────────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None,
        prefix: str,
        suffix: str,
        exclude_columns: Sequence[str] = (),
        propagate_nulls: bool = True,
        exist_policy: str = "raise",
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._prefix = prefix
        self._suffix = suffix
        self._propagate_nulls = propagate_nulls

        check_column_exist_policy(exist_policy)
        self._exist_policy = exist_policy

        check_sklearn()
        self._scaler = sklearn.preprocessing.MinMaxScaler(**kwargs)
        self._kwargs = kwargs

    def __repr__(self) -> str:
        args = repr_mapping_line(
            {
                "columns": self._columns,
                "prefix": self._prefix,
                "suffix": self._suffix,
                "exclude_columns": self._exclude_columns,
                "propagate_nulls": self._propagate_nulls,
                "exist_policy": self._exist_policy,
                "missing_policy": self._missing_policy,
            }
        )
        return f"{self.__class__.__qualname__}({args}{str_kwargs(self._kwargs)})"

    def _fit(self, frame: pl.DataFrame) -> None:
        logger.info(
            f"Fitting the min/max scaling parameters of {len(self.find_columns(frame)):,} "
            "columns..."
        )
        columns = self.find_common_columns(frame)
        self._scaler.fit(frame.select(columns).to_numpy())

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        self._check_output_columns(frame)
        logger.info(
            f"Applying the min/max scaling transformation on {len(self.find_columns(frame)):,} "
            f"columns | prefix={self._prefix!r} | suffix={self._suffix!r}"
        )
        columns = self.find_common_columns(frame)
        data = frame.select(columns)

        x = self._scaler.transform(data.to_numpy())
        data_scaled = pl.from_numpy(x, schema=data.columns)
        if self._propagate_nulls:
            data_scaled = propagate_nulls(data_scaled, data)
        return frame.with_columns(
            data_scaled.rename(lambda col: f"{self._prefix}{col}{self._suffix}")
        )

    def _check_output_columns(self, frame: pl.DataFrame) -> None:
        r"""Check if the output columns already exist.

        Args:
            frame: The input DataFrame to check.
        """
        check_existing_columns(
            frame,
            columns=[f"{self._prefix}{col}{self._suffix}" for col in self.find_columns(frame)],
            exist_policy=self._exist_policy,
        )



r"""Contain utility functions to compute hash of objects."""

from __future__ import annotations

__all__ = ["str_to_sha256"]

import hashlib


def str_to_sha256(string: str) -> str:
    r"""Generate the SHA-256 hash of a string.

    Args:
        string: The string to hash.

    Returns:
        The SHA-256 hash.

    Example usage:

    ```pycon

    >>> from grizz.utils.hashing import str_to_sha256
    >>> str_to_sha256("bears are funny")
    c97afc5c7f1b598c9f68dc2d6e323b2dd2eaaa31d3a07c98059de6079cbd30e0

    ```
    """
    return hashlib.sha256(str(string).encode("utf-8")).hexdigest()



r"""Contain utility functions for datetime and date objects."""

from __future__ import annotations

__all__ = ["find_end_datetime", "to_datetime"]


from datetime import date, datetime, timedelta, timezone

from grizz.utils.interval import interval_to_timedelta


def find_end_datetime(start: datetime | date, interval: str | timedelta, periods: int) -> datetime:
    r"""Find the upper bound of the datetime range from the lower bound
    of the datetime range, the interval, and the number of periods.

    Args:
        start: The lower bound of the datetime range.
        interval: The interval of the range periods, specified as a
            Python timedelta object or using the Polars duration
            string language.
        periods: The number of periods after the start.

    Returns:
        The upper bound of the datetime range.

    Notes:
        ``interval`` is created according to the following string
            language:

            - 1ns (1 nanosecond)
            - 1us (1 microsecond)
            - 1ms (1 millisecond)
            - 1s (1 second)
            - 1m (1 minute)
            - 1h (1 hour)
            - 1d (1 calendar day)
            - 1w (1 calendar week)

    Example usage:

    ```pycon

    >>> from datetime import timedelta, datetime, timezone
    >>> from grizz.utils.datetime import find_end_datetime
    >>> find_end_datetime(
    ...     start=datetime(year=2020, month=5, day=12, hour=4, tzinfo=timezone.utc),
    ...     interval=timedelta(hours=1),
    ...     periods=42,
    ... )
    datetime.datetime(2020, 5, 13, 22, 0, tzinfo=datetime.timezone.utc)

    ```
    """
    start = to_datetime(start)
    if isinstance(interval, str):
        interval = interval_to_timedelta(interval)
    return start + interval * periods


def to_datetime(dt: datetime | date) -> datetime:
    r"""Convert a ``date`` object to a ``datetime`` object.

    Args:
        dt: The ``date`` object to convert.

    Returns:
        The ``datetime`` object.

    Example usage:

    ```pycon

    >>> from datetime import datetime, date, timezone
    >>> from grizz.utils.datetime import to_datetime
    >>> to_datetime(datetime(year=2020, month=5, day=12, hour=4, tzinfo=timezone.utc))
    datetime.datetime(2020, 5, 12, 4, 0, tzinfo=datetime.timezone.utc)
    >>> to_datetime(date(year=2020, month=5, day=12))
    datetime.datetime(2020, 5, 12, 0, 0, tzinfo=datetime.timezone.utc)

    ```
    """
    if isinstance(dt, datetime):
        return dt
    return datetime(dt.year, dt.month, dt.day, tzinfo=timezone.utc)



r"""Contain transformers to fill values."""

from __future__ import annotations

__all__ = ["FillNanTransformer", "FillNullTransformer"]

import logging
from typing import TYPE_CHECKING, Any

import polars as pl
import polars.selectors as cs

from grizz.transformer.columns import BaseInNTransformer

if TYPE_CHECKING:
    from collections.abc import Sequence


logger = logging.getLogger(__name__)


class FillNanTransformer(BaseInNTransformer):
    r"""Implement a transformer to fill NaN values.

    Args:
        columns: The columns to prepare. If ``None``, it processes all
            the columns of type string.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``fill_nan``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import FillNan
    >>> transformer = FillNan(columns=["col1", "col4"], value=100)
    >>> transformer
    FillNanTransformer(columns=('col1', 'col4'), exclude_columns=(), missing_policy='raise', value=100)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, None],
    ...         "col2": [1.2, 2.2, 3.2, 4.2, float("nan")],
    ...         "col3": ["a", "b", "c", "d", None],
    ...         "col4": [1.2, float("nan"), 3.2, None, 5.2],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ f64  ┆ str  ┆ f64  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1.2  ┆ a    ┆ 1.2  │
    │ 2    ┆ 2.2  ┆ b    ┆ NaN  │
    │ 3    ┆ 3.2  ┆ c    ┆ 3.2  │
    │ 4    ┆ 4.2  ┆ d    ┆ null │
    │ null ┆ NaN  ┆ null ┆ 5.2  │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬───────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4  │
    │ ---  ┆ ---  ┆ ---  ┆ ---   │
    │ i64  ┆ f64  ┆ str  ┆ f64   │
    ╞══════╪══════╪══════╪═══════╡
    │ 1    ┆ 1.2  ┆ a    ┆ 1.2   │
    │ 2    ┆ 2.2  ┆ b    ┆ 100.0 │
    │ 3    ┆ 3.2  ┆ c    ┆ 3.2   │
    │ 4    ┆ 4.2  ┆ d    ┆ null  │
    │ null ┆ NaN  ┆ null ┆ 5.2   │
    └──────┴──────┴──────┴───────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._kwargs = kwargs

    def get_args(self) -> dict:
        return super().get_args() | self._kwargs

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Filling NaN values of {len(self.find_columns(frame)):,} columns...")
        columns = self.find_common_columns(frame)
        return frame.with_columns(
            frame.select((cs.by_name(columns) & cs.float()).fill_nan(**self._kwargs))
        )


class FillNullTransformer(BaseInNTransformer):
    r"""Implement a transformer to fill null values.

    Args:
        columns: The columns to prepare. If ``None``, it processes all
            the columns of type string.
        exclude_columns: The columns to exclude from the input
            ``columns``. If any column is not found, it will be ignored
            during the filtering process.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.
        **kwargs: The keyword arguments for ``fill_null``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import FillNull
    >>> transformer = FillNull(columns=["col1", "col4"], value=100)
    >>> transformer
    FillNullTransformer(columns=('col1', 'col4'), exclude_columns=(), missing_policy='raise', value=100)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, None],
    ...         "col2": [1.2, 2.2, 3.2, 4.2, None],
    ...         "col3": ["a", "b", "c", "d", None],
    ...         "col4": [1.2, float("nan"), 3.2, None, 5.2],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ f64  ┆ str  ┆ f64  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1.2  ┆ a    ┆ 1.2  │
    │ 2    ┆ 2.2  ┆ b    ┆ NaN  │
    │ 3    ┆ 3.2  ┆ c    ┆ 3.2  │
    │ 4    ┆ 4.2  ┆ d    ┆ null │
    │ null ┆ null ┆ null ┆ 5.2  │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬───────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4  │
    │ ---  ┆ ---  ┆ ---  ┆ ---   │
    │ i64  ┆ f64  ┆ str  ┆ f64   │
    ╞══════╪══════╪══════╪═══════╡
    │ 1    ┆ 1.2  ┆ a    ┆ 1.2   │
    │ 2    ┆ 2.2  ┆ b    ┆ NaN   │
    │ 3    ┆ 3.2  ┆ c    ┆ 3.2   │
    │ 4    ┆ 4.2  ┆ d    ┆ 100.0 │
    │ 100  ┆ null ┆ null ┆ 5.2   │
    └──────┴──────┴──────┴───────┘

    ```
    """

    def __init__(
        self,
        columns: Sequence[str] | None = None,
        exclude_columns: Sequence[str] = (),
        missing_policy: str = "raise",
        **kwargs: Any,
    ) -> None:
        super().__init__(
            columns=columns,
            exclude_columns=exclude_columns,
            missing_policy=missing_policy,
        )
        self._kwargs = kwargs

    def get_args(self) -> dict:
        return super().get_args() | self._kwargs

    def _fit(self, frame: pl.DataFrame) -> None:  # noqa: ARG002
        logger.info(
            f"Skipping '{self.__class__.__qualname__}.fit' as there are no parameters "
            f"available to fit"
        )

    def _transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        logger.info(f"Filling NaN values of {len(self.find_columns(frame)):,} columns...")
        columns = self.find_common_columns(frame)
        return frame.with_columns(frame.select(cs.by_name(columns).fill_null(**self._kwargs)))



r"""Contain the base class to implement a ``polars.DataFrame``
transformer."""

from __future__ import annotations

__all__ = [
    "BaseTransformer",
    "is_transformer_config",
    "setup_transformer",
]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import polars as pl

logger = logging.getLogger(__name__)


class BaseTransformer(ABC, metaclass=AbstractFactory):
    r"""Define the base class to transform a ``polars.DataFrame``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import Cast
    >>> transformer = Cast(columns=["col1", "col3"], dtype=pl.Int32)
    >>> transformer
    CastTransformer(columns=('col1', 'col3'), dtype=Int32, exclude_columns=(), missing_policy='raise')
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["1", "2", "3", "4", "5"],
    ...         "col4": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> frame
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i64  ┆ str  ┆ str  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘
    >>> out = transformer.transform(frame)
    >>> out
    shape: (5, 4)
    ┌──────┬──────┬──────┬──────┐
    │ col1 ┆ col2 ┆ col3 ┆ col4 │
    │ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ i32  ┆ str  ┆ i32  ┆ str  │
    ╞══════╪══════╪══════╪══════╡
    │ 1    ┆ 1    ┆ 1    ┆ a    │
    │ 2    ┆ 2    ┆ 2    ┆ b    │
    │ 3    ┆ 3    ┆ 3    ┆ c    │
    │ 4    ┆ 4    ┆ 4    ┆ d    │
    │ 5    ┆ 5    ┆ 5    ┆ e    │
    └──────┴──────┴──────┴──────┘

    ```
    """

    @abstractmethod
    def fit(self, frame: pl.DataFrame) -> None:
        r"""Fit to the data in the ``polars.DataFrame``.

        Args:
            frame: The ``polars.DataFrame`` to fit.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from grizz.transformer import Cast
        >>> transformer = Cast(columns=["col1", "col3"], dtype=pl.Int32)
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> frame
        shape: (5, 4)
        ┌──────┬──────┬──────┬──────┐
        │ col1 ┆ col2 ┆ col3 ┆ col4 │
        │ ---  ┆ ---  ┆ ---  ┆ ---  │
        │ i64  ┆ str  ┆ str  ┆ str  │
        ╞══════╪══════╪══════╪══════╡
        │ 1    ┆ 1    ┆ 1    ┆ a    │
        │ 2    ┆ 2    ┆ 2    ┆ b    │
        │ 3    ┆ 3    ┆ 3    ┆ c    │
        │ 4    ┆ 4    ┆ 4    ┆ d    │
        │ 5    ┆ 5    ┆ 5    ┆ e    │
        └──────┴──────┴──────┴──────┘
        >>> transformer.fit(frame)
        >>> out = transformer.transform(frame)
        >>> out
        shape: (5, 4)
        ┌──────┬──────┬──────┬──────┐
        │ col1 ┆ col2 ┆ col3 ┆ col4 │
        │ ---  ┆ ---  ┆ ---  ┆ ---  │
        │ i32  ┆ str  ┆ i32  ┆ str  │
        ╞══════╪══════╪══════╪══════╡
        │ 1    ┆ 1    ┆ 1    ┆ a    │
        │ 2    ┆ 2    ┆ 2    ┆ b    │
        │ 3    ┆ 3    ┆ 3    ┆ c    │
        │ 4    ┆ 4    ┆ 4    ┆ d    │
        │ 5    ┆ 5    ┆ 5    ┆ e    │
        └──────┴──────┴──────┴──────┘

        ```
        """

    @abstractmethod
    def fit_transform(self, frame: pl.DataFrame) -> None:
        r"""Fit to the data, then transform it.

        Args:
            frame: The ``polars.DataFrame`` to fit.

        Returns:
            The transformed DataFrame.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from grizz.transformer import Cast
        >>> transformer = Cast(columns=["col1", "col3"], dtype=pl.Int32)
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> frame
        shape: (5, 4)
        ┌──────┬──────┬──────┬──────┐
        │ col1 ┆ col2 ┆ col3 ┆ col4 │
        │ ---  ┆ ---  ┆ ---  ┆ ---  │
        │ i64  ┆ str  ┆ str  ┆ str  │
        ╞══════╪══════╪══════╪══════╡
        │ 1    ┆ 1    ┆ 1    ┆ a    │
        │ 2    ┆ 2    ┆ 2    ┆ b    │
        │ 3    ┆ 3    ┆ 3    ┆ c    │
        │ 4    ┆ 4    ┆ 4    ┆ d    │
        │ 5    ┆ 5    ┆ 5    ┆ e    │
        └──────┴──────┴──────┴──────┘
        >>> out = transformer.fit_transform(frame)
        >>> out
        shape: (5, 4)
        ┌──────┬──────┬──────┬──────┐
        │ col1 ┆ col2 ┆ col3 ┆ col4 │
        │ ---  ┆ ---  ┆ ---  ┆ ---  │
        │ i32  ┆ str  ┆ i32  ┆ str  │
        ╞══════╪══════╪══════╪══════╡
        │ 1    ┆ 1    ┆ 1    ┆ a    │
        │ 2    ┆ 2    ┆ 2    ┆ b    │
        │ 3    ┆ 3    ┆ 3    ┆ c    │
        │ 4    ┆ 4    ┆ 4    ┆ d    │
        │ 5    ┆ 5    ┆ 5    ┆ e    │
        └──────┴──────┴──────┴──────┘

        ```
        """

    @abstractmethod
    def transform(self, frame: pl.DataFrame) -> pl.DataFrame:
        r"""Transform the data in the ``polars.DataFrame``.

        Args:
            frame: The ``polars.DataFrame`` to transform.

        Returns:
            The transformed DataFrame.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from grizz.transformer import Cast
        >>> transformer = Cast(columns=["col1", "col3"], dtype=pl.Int32)
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["1", "2", "3", "4", "5"],
        ...         "col4": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> frame
        shape: (5, 4)
        ┌──────┬──────┬──────┬──────┐
        │ col1 ┆ col2 ┆ col3 ┆ col4 │
        │ ---  ┆ ---  ┆ ---  ┆ ---  │
        │ i64  ┆ str  ┆ str  ┆ str  │
        ╞══════╪══════╪══════╪══════╡
        │ 1    ┆ 1    ┆ 1    ┆ a    │
        │ 2    ┆ 2    ┆ 2    ┆ b    │
        │ 3    ┆ 3    ┆ 3    ┆ c    │
        │ 4    ┆ 4    ┆ 4    ┆ d    │
        │ 5    ┆ 5    ┆ 5    ┆ e    │
        └──────┴──────┴──────┴──────┘
        >>> out = transformer.transform(frame)
        >>> out
        shape: (5, 4)
        ┌──────┬──────┬──────┬──────┐
        │ col1 ┆ col2 ┆ col3 ┆ col4 │
        │ ---  ┆ ---  ┆ ---  ┆ ---  │
        │ i32  ┆ str  ┆ i32  ┆ str  │
        ╞══════╪══════╪══════╪══════╡
        │ 1    ┆ 1    ┆ 1    ┆ a    │
        │ 2    ┆ 2    ┆ 2    ┆ b    │
        │ 3    ┆ 3    ┆ 3    ┆ c    │
        │ 4    ┆ 4    ┆ 4    ┆ d    │
        │ 5    ┆ 5    ┆ 5    ┆ e    │
        └──────┴──────┴──────┴──────┘

        ```
        """


def is_transformer_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseTransformer``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseTransformer`` object.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import is_transformer_config
    >>> is_transformer_config(
    ...     {
    ...         "_target_": "grizz.transformer.Cast",
    ...         "columns": ("col1", "col3"),
    ...         "dtype": pl.Int32,
    ...     }
    ... )
    True

    ```
    """
    return is_object_config(config, BaseTransformer)


def setup_transformer(
    transformer: BaseTransformer | dict,
) -> BaseTransformer:
    r"""Set up a ``polars.DataFrame`` transformer.

    The transformer is instantiated from its configuration
    by using the ``BaseTransformer`` factory function.

    Args:
        transformer: Specifies a ``polars.DataFrame`` transformer or
            its configuration.

    Returns:
        An instantiated transformer.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.transformer import setup_transformer
    >>> transformer = setup_transformer(
    ...     {
    ...         "_target_": "grizz.transformer.Cast",
    ...         "columns": ("col1", "col3"),
    ...         "dtype": pl.Int32,
    ...     }
    ... )
    >>> transformer
    CastTransformer(columns=('col1', 'col3'), dtype=Int32, exclude_columns=(), missing_policy='raise')

    ```
    """
    if isinstance(transformer, dict):
        logger.info("Initializing a DataFrame transformer from its configuration... ")
        transformer = BaseTransformer.factory(**transformer)
    if not isinstance(transformer, BaseTransformer):
        logger.warning(f"transformer is not a `BaseTransformer` (received: {type(transformer)})")
    return transformer



r"""Contain utility functions to do temporal transformations."""

from __future__ import annotations

__all__ = ["compute_temporal_stats", "to_step_names", "to_temporal_frames"]


import polars as pl

from grizz.utils.interval import interval_to_strftime_format


def compute_temporal_stats(
    frame: pl.DataFrame,
    column: str,
    temporal_column: str,
    period: str,
) -> pl.DataFrame:
    r"""Return a DataFrame with stats for each temporal window.

    Args:
        frame: The DataFrame to analyze.
        column: The column to analyze.
        temporal_column: The temporal column used to create the temporal
            DataFrames.
        period: The temporal period e.g. monthly or daily.

    Returns:
        A DataFrame with stats for each temporal window.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from grizz.utils.temporal import compute_temporal_stats
    >>> stats = compute_temporal_stats(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col": [1.2, 4.2, 0.0, 1.0, 4.2, 42.0],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col": pl.Float64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     column="col",
    ...     temporal_column="datetime",
    ...     period="1mo",
    ... )
    >>> stats
    shape: (4, 16)
    ┌─────────────────────────┬───────┬─────────┬──────┬───┬──────┬──────┬──────┬──────┐
    │ step                    ┆ count ┆ nunique ┆ mean ┆ … ┆ q90  ┆ q95  ┆ q99  ┆ max  │
    │ ---                     ┆ ---   ┆ ---     ┆ ---  ┆   ┆ ---  ┆ ---  ┆ ---  ┆ ---  │
    │ datetime[μs, UTC]       ┆ i64   ┆ i64     ┆ f64  ┆   ┆ f64  ┆ f64  ┆ f64  ┆ f64  │
    ╞═════════════════════════╪═══════╪═════════╪══════╪═══╪══════╪══════╪══════╪══════╡
    │ 2020-01-01 00:00:00 UTC ┆ 3     ┆ 3       ┆ 1.8  ┆ … ┆ 4.2  ┆ 4.2  ┆ 4.2  ┆ 4.2  │
    │ 2020-02-01 00:00:00 UTC ┆ 1     ┆ 1       ┆ 1.0  ┆ … ┆ 1.0  ┆ 1.0  ┆ 1.0  ┆ 1.0  │
    │ 2020-03-01 00:00:00 UTC ┆ 1     ┆ 1       ┆ 4.2  ┆ … ┆ 4.2  ┆ 4.2  ┆ 4.2  ┆ 4.2  │
    │ 2020-04-01 00:00:00 UTC ┆ 1     ┆ 1       ┆ 42.0 ┆ … ┆ 42.0 ┆ 42.0 ┆ 42.0 ┆ 42.0 │
    └─────────────────────────┴───────┴─────────┴──────┴───┴──────┴──────┴──────┴──────┘

    ```
    """
    return (
        frame.select(column, temporal_column)
        .sort(temporal_column)
        .group_by_dynamic(temporal_column, every=period)
        .agg(
            pl.col(column).len().cast(pl.Int64).alias("count"),
            pl.col(column).n_unique().cast(pl.Int64).alias("nunique"),
            pl.col(column).mean().cast(pl.Float64).alias("mean"),
            pl.col(column).std().cast(pl.Float64).alias("std"),
            pl.col(column).min().cast(pl.Float64).alias("min"),
            pl.col(column).quantile(0.01).cast(pl.Float64).alias("q01"),
            pl.col(column).quantile(0.05).cast(pl.Float64).alias("q05"),
            pl.col(column).quantile(0.1).cast(pl.Float64).alias("q10"),
            pl.col(column).quantile(0.25).cast(pl.Float64).alias("q25"),
            pl.col(column).median().cast(pl.Float64).alias("median"),
            pl.col(column).quantile(0.75).cast(pl.Float64).alias("q75"),
            pl.col(column).quantile(0.9).cast(pl.Float64).alias("q90"),
            pl.col(column).quantile(0.95).cast(pl.Float64).alias("q95"),
            pl.col(column).quantile(0.99).cast(pl.Float64).alias("q99"),
            pl.col(column).max().cast(pl.Float64).alias("max"),
        )
        .rename({temporal_column: "step"})
    )


def to_temporal_frames(
    frame: pl.DataFrame,
    temporal_column: str,
    period: str,
) -> tuple[list[pl.DataFrame], list[str]]:
    r"""Return a list of temporal DataFrames and the associated time
    steps.

    Args:
        frame: The DataFrame to analyze.
        temporal_column: The temporal column used to create the temporal
            DataFrames.
        period: The temporal period e.g. monthly or daily.

    Returns:
        A tuple with the counts and the temporal steps.

    Example usage:

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from grizz.utils.temporal import to_temporal_frames
    >>> frames, steps = to_temporal_frames(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col1": [None, float("nan"), 0.0, 1.0, 4.2, 42.0],
    ...             "col2": [None, 1, 0, None, 2, 3],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col1": pl.Float64,
    ...             "col2": pl.Int64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     ),
    ...     temporal_column="datetime",
    ...     period="1mo",
    ... )
    >>> frames
    [shape: (3, 3)
    ┌──────┬──────┬─────────────────────────┐
    │ col1 ┆ col2 ┆ datetime                │
    │ ---  ┆ ---  ┆ ---                     │
    │ f64  ┆ i64  ┆ datetime[μs, UTC]       │
    ╞══════╪══════╪═════════════════════════╡
    │ null ┆ null ┆ 2020-01-03 00:00:00 UTC │
    │ NaN  ┆ 1    ┆ 2020-01-04 00:00:00 UTC │
    │ 0.0  ┆ 0    ┆ 2020-01-05 00:00:00 UTC │
    └──────┴──────┴─────────────────────────┘, shape: (1, 3)
    ┌──────┬──────┬─────────────────────────┐
    │ col1 ┆ col2 ┆ datetime                │
    │ ---  ┆ ---  ┆ ---                     │
    │ f64  ┆ i64  ┆ datetime[μs, UTC]       │
    ╞══════╪══════╪═════════════════════════╡
    │ 1.0  ┆ null ┆ 2020-02-03 00:00:00 UTC │
    └──────┴──────┴─────────────────────────┘, shape: (1, 3)
    ┌──────┬──────┬─────────────────────────┐
    │ col1 ┆ col2 ┆ datetime                │
    │ ---  ┆ ---  ┆ ---                     │
    │ f64  ┆ i64  ┆ datetime[μs, UTC]       │
    ╞══════╪══════╪═════════════════════════╡
    │ 4.2  ┆ 2    ┆ 2020-03-03 00:00:00 UTC │
    └──────┴──────┴─────────────────────────┘, shape: (1, 3)
    ┌──────┬──────┬─────────────────────────┐
    │ col1 ┆ col2 ┆ datetime                │
    │ ---  ┆ ---  ┆ ---                     │
    │ f64  ┆ i64  ┆ datetime[μs, UTC]       │
    ╞══════╪══════╪═════════════════════════╡
    │ 42.0 ┆ 3    ┆ 2020-04-03 00:00:00 UTC │
    └──────┴──────┴─────────────────────────┘]
    >>> steps
    ['2020-01', '2020-02', '2020-03', '2020-04']

    ```
    """
    if frame.is_empty():
        return [], []

    groups = frame.sort(temporal_column).group_by_dynamic(temporal_column, every=period)
    steps = to_step_names(groups=groups, period=period)
    frames = [frame for _, frame in groups]
    return frames, steps


def to_step_names(groups: pl.GroupBy, period: str) -> list[str]:
    r"""Return the name of each step.

    Args:
        groups: The grouped DataFrame by step.
        period: The temporal period e.g. monthly or daily.

    Returns:
        A list that contains the name of each step.

    ```pycon

    >>> from datetime import datetime, timezone
    >>> import polars as pl
    >>> from grizz.utils.temporal import to_step_names
    >>> groups = (
    ...     pl.DataFrame(
    ...         {
    ...             "col": [0.0, 1.0, 2.0, 3.0, 4.0, 5.0],
    ...             "datetime": [
    ...                 datetime(year=2020, month=1, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=4, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=1, day=5, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=2, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=3, day=3, tzinfo=timezone.utc),
    ...                 datetime(year=2020, month=4, day=3, tzinfo=timezone.utc),
    ...             ],
    ...         },
    ...         schema={
    ...             "col": pl.Float64,
    ...             "datetime": pl.Datetime(time_unit="us", time_zone="UTC"),
    ...         },
    ...     )
    ...     .sort("datetime")
    ...     .group_by_dynamic("datetime", every="1mo")
    ... )
    >>> steps = to_step_names(groups=groups, period="1mo")
    >>> steps
    ['2020-01', '2020-02', '2020-03', '2020-04']

    ```
    """
    format_dt = interval_to_strftime_format(period)
    return [name[0].strftime(format_dt) for name, _ in groups]



r"""Contain the base class to implement a DataFrame exporter."""

from __future__ import annotations

__all__ = ["BaseExporter", "is_exporter_config", "setup_exporter"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any

from coola.equality.comparators import BaseEqualityComparator
from coola.equality.handlers import EqualNanHandler, SameObjectHandler, SameTypeHandler
from coola.equality.testers import EqualityTester
from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    import polars as pl
    from coola.equality import EqualityConfig

logger = logging.getLogger(__name__)


class BaseExporter(ABC, metaclass=AbstractFactory):
    r"""Define the base class to implement a DataFrame exporter.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.exporter import ParquetExporter
    >>> exporter = ParquetExporter(path="/path/to/frame.parquet")
    >>> exporter
    ParquetExporter(path=/path/to/frame.parquet)
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a", "b", "c", "d", "e"],
    ...     }
    ... )
    >>> exporter.export(frame)  # doctest: +SKIP

    ```
    """

    @abstractmethod
    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        r"""Indicate if two exporter objects are equal or not.

        Args:
            other: The other object to compare.
            equal_nan: Whether to compare NaN's as equal. If ``True``,
                NaN's in both objects will be considered equal.

        Returns:
            ``True`` if the two exporters are equal, otherwise ``False``.

        Example usage:

        ```pycon

        >>> import numpy as np
        >>> from grizz.exporter import CsvExporter
        >>> obj1 = CsvExporter(path="/path/to/frame.csv")
        >>> obj2 = CsvExporter(path="/path/to/frame.csv")
        >>> obj3 = CsvExporter(path="/path/to/frame2.csv")
        >>> obj1.equal(obj2)
        True
        >>> obj1.equal(obj3)
        False

        ```
        """

    @abstractmethod
    def export(self, frame: pl.DataFrame) -> None:
        r"""Export a DataFrame.

        Args:
            frame: The DataFrame to export.

        Example usage:

        ```pycon

        >>> import polars as pl
        >>> from grizz.exporter import ParquetExporter
        >>> exporter = ParquetExporter(path="/path/to/frame.parquet")
        >>> frame = pl.DataFrame(
        ...     {
        ...         "col1": [1, 2, 3, 4, 5],
        ...         "col2": ["1", "2", "3", "4", "5"],
        ...         "col3": ["a", "b", "c", "d", "e"],
        ...     }
        ... )
        >>> exporter.export(frame)  # doctest: +SKIP

        ```
        """


def is_exporter_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseExporter``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: The configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration
            for a ``BaseExporter`` object.

    Example usage:

    ```pycon

    >>> from grizz.exporter import is_exporter_config
    >>> is_exporter_config(
    ...     {"_target_": "grizz.exporter.ParquetExporter", "path": "/path/to/data.parquet"}
    ... )
    True

    ```
    """
    return is_object_config(config, BaseExporter)


def setup_exporter(
    exporter: BaseExporter | dict,
) -> BaseExporter:
    r"""Set up an exporter.

    The exporter is instantiated from its configuration
    by using the ``BaseExporter`` factory function.

    Args:
        exporter: A exporter or its configuration.

    Returns:
        An instantiated exporter.

    Example usage:

    ```pycon

    >>> from grizz.exporter import setup_exporter
    >>> exporter = setup_exporter(
    ...     {"_target_": "grizz.exporter.ParquetExporter", "path": "/path/to/data.parquet"}
    ... )
    >>> exporter
    ParquetExporter(path=/path/to/data.parquet)

    ```
    """
    if isinstance(exporter, dict):
        logger.info("Initializing an exporter from its configuration... ")
        exporter = BaseExporter.factory(**exporter)
    if not isinstance(exporter, BaseExporter):
        logger.warning(f"exporter is not a `BaseExporter` (received: {type(exporter)})")
    return exporter


class ExporterEqualityComparator(BaseEqualityComparator[BaseExporter]):
    r"""Implement an equality comparator for ``BaseExporter``
    objects."""

    def __init__(self) -> None:
        self._handler = SameObjectHandler()
        self._handler.chain(SameTypeHandler()).chain(EqualNanHandler())

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def clone(self) -> ExporterEqualityComparator:
        return self.__class__()

    def equal(self, actual: BaseExporter, expected: Any, config: EqualityConfig) -> bool:
        return self._handler.handle(actual, expected, config=config)


if not EqualityTester.has_comparator(BaseExporter):  # pragma: no cover
    EqualityTester.add_comparator(BaseExporter, ExporterEqualityComparator())



r"""Contain the implementation of a simple ingestor."""

from __future__ import annotations

__all__ = ["Ingestor"]


from typing import TYPE_CHECKING, Any

from coola import objects_are_equal

from grizz.ingestor.base import BaseIngestor

if TYPE_CHECKING:
    import polars as pl


class Ingestor(BaseIngestor):
    r"""Implement a simple DataFrame ingestor.

    Args:
        frame: The DataFrame to ingest.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.ingestor import Ingestor
    >>> ingestor = Ingestor(
    ...     frame=pl.DataFrame(
    ...         {
    ...             "col1": [1, 2, 3, 4, 5],
    ...             "col2": ["1", "2", "3", "4", "5"],
    ...             "col3": ["1", "2", "3", "4", "5"],
    ...             "col4": ["a", "b", "c", "d", "e"],
    ...         }
    ...     )
    ... )
    >>> ingestor
    Ingestor(shape=(5, 4))
    >>> frame = ingestor.ingest()

    ```
    """

    def __init__(self, frame: pl.DataFrame) -> None:
        self._frame = frame

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(shape={self._frame.shape})"

    def equal(self, other: Any, equal_nan: bool = False) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return objects_are_equal(self._frame, other._frame, equal_nan=equal_nan)

    def ingest(self) -> pl.DataFrame:
        return self._frame



r"""Contain interval utility functions."""

from __future__ import annotations

__all__ = [
    "find_time_unit",
    "interval_to_strftime_format",
    "interval_to_timedelta",
    "time_unit_to_strftime_format",
]

import re
from datetime import timedelta

STRFTIME_FORMAT = {
    "ns": "%Y-%m-%d %H:%M:%S.%f",
    "us": "%Y-%m-%d %H:%M:%S.%f",
    "ms": "%Y-%m-%d %H:%M:%S.%f",
    "s": "%Y-%m-%d %H:%M:%S",
    "m": "%Y-%m-%d %H:%M",
    "h": "%Y-%m-%d %H:%M",
    "d": "%Y-%m-%d",
    "w": "%Y week %W",
    "mo": "%Y-%m",
    "q": "%Y-%m",
    "y": "%Y",
}

TIME_UNIT_TO_INTERVAL_REGEX = {
    "ns": "[0-9]+ns([0-9]|$)",
    "us": "[0-9]+us([0-9]|$)",
    "ms": "[0-9]+ms([0-9]|$)",
    "s": "[0-9]+s([0-9]|$)",
    "m": "[0-9]+m([0-9]|$)",
    "h": "[0-9]+h([0-9]|$)",
    "d": "[0-9]+d([0-9]|$)",
    "w": "[0-9]+w([0-9]|$)",
    "mo": "[0-9]+mo([0-9]|$)",
    "q": "[0-9]+q([0-9]|$)",
    "y": "[0-9]+y([0-9]|$)",
}


def find_time_unit(interval: str) -> str:
    r"""Find the time unit associated to a ``polars`` interval.

    Args:
        interval: The ``polars`` interval to analyze.

    Returns:
        The found time unit.

    Raises:
        RuntimeError: if no valid time unit can be found.

    Example usage:

    ```pycon

    >>> from grizz.utils.interval import find_time_unit
    >>> find_time_unit("3d12h4m")
    m
    >>> find_time_unit("3y5mo")
    mo

    ```
    """
    for unit, regex in TIME_UNIT_TO_INTERVAL_REGEX.items():
        if re.compile(regex).search(interval) is not None:
            return unit

    msg = f"could not find the time unit of {interval}"
    raise RuntimeError(msg)


def interval_to_strftime_format(interval: str) -> str:
    r"""Return the default strftime format for a given interval.

    Args:
        interval: The ``polars`` interval to analyze.

    Returns:
        The default strftime format.

    Example usage:

    ```pycon

    >>> from grizz.utils.interval import interval_to_strftime_format
    >>> interval_to_strftime_format("1h")
    %Y-%m-%d %H:%M
    >>> interval_to_strftime_format("3y1mo")
    %Y-%m

    ```
    """
    return time_unit_to_strftime_format(time_unit=find_time_unit(interval))


def interval_to_timedelta(interval: str) -> timedelta:
    r"""Convert a interval to a timedelta object.

    Args:
        interval: The input interval.

    Returns:
        The timedelta object generated from the interval.

    Example usage:

    ```pycon

    >>> from grizz.utils.interval import interval_to_timedelta
    >>> interval_to_timedelta("5d1h42m")
    datetime.timedelta(days=5, seconds=6120)

    ```
    """

    def extract(regex: str, interval: str) -> float | None:
        res = re.compile(regex).search(interval)
        if res is None:
            return 0.0
        return float(re.compile("[0-9]+").search(res[0])[0])

    days = extract(TIME_UNIT_TO_INTERVAL_REGEX["w"], interval)
    microseconds = extract(TIME_UNIT_TO_INTERVAL_REGEX["ns"], interval) * 0.001
    return timedelta(
        days=extract(TIME_UNIT_TO_INTERVAL_REGEX["d"], interval) + 7 * days,
        hours=extract(TIME_UNIT_TO_INTERVAL_REGEX["h"], interval),
        minutes=extract(TIME_UNIT_TO_INTERVAL_REGEX["m"], interval),
        seconds=extract(TIME_UNIT_TO_INTERVAL_REGEX["s"], interval),
        milliseconds=extract(TIME_UNIT_TO_INTERVAL_REGEX["ms"], interval),
        microseconds=extract(TIME_UNIT_TO_INTERVAL_REGEX["us"], interval) + microseconds,
    )


def time_unit_to_strftime_format(time_unit: str) -> str:
    r"""Return the default strftime format for a given time unit.

    Args:
        time_unit: The time unit.

    Returns:
        The default strftime format.

    Example usage:

    ```pycon

    >>> from grizz.utils.interval import time_unit_to_strftime_format
    >>> time_unit_to_strftime_format("h")
    %Y-%m-%d %H:%M
    >>> time_unit_to_strftime_format("mo")
    %Y-%m

    ```
    """
    template = STRFTIME_FORMAT.get(time_unit.lower(), None)
    if template is None:
        msg = (
            f"Incorrect time unit {time_unit}. The valid time units are: "
            f"{list(STRFTIME_FORMAT.keys())}"
        )
        raise RuntimeError(msg)
    return template



r"""Contain custom exceptions."""

from __future__ import annotations

__all__ = [
    "ColumnExistsError",
    "ColumnExistsWarning",
    "ColumnNotFoundError",
    "ColumnNotFoundWarning",
    "DataFrameNotFoundError",
]


class ColumnExistsError(RuntimeError):
    r"""Raised when trying to create a column which already exists."""


class ColumnExistsWarning(RuntimeWarning):
    r"""Raised when trying to create a column which already exists."""


class ColumnNotFoundError(RuntimeError):
    r"""Raised when a column is requested but does not exist."""


class ColumnNotFoundWarning(RuntimeWarning):
    r"""Raised when a column is requested but does not exist."""


class DataFrameNotFoundError(RuntimeError):
    r"""Raised when a DataFrame is requested but does not exist."""



r"""Contain utility functions to transform data with NaNs."""

from __future__ import annotations

__all__ = ["remove_nan", "sortnan"]

import math
from collections.abc import Iterable, Sequence
from typing import Any, TypeVar

T = TypeVar("T", bound=Sequence)


def remove_nan(data: T) -> T:
    r"""Remove the NaN values from the input sequence.

    Args:
        data: The input sequence.

    Returns:
        The input sequence without NaN values.

    Example usage:

    ```pycon

    >>> from grizz.utils.nan import remove_nan
    >>> data = [float("nan"), float("-inf"), -2, 1.2]
    >>> remove_nan(data)
    [-inf, -2, 1.2]

    ```
    """
    return type(data)([x for x in data if not isinstance(x, (float, int)) or not math.isnan(x)])


def sortnan(iterable: Iterable[bool | float], /, *, reverse: bool = False) -> list[bool | float]:
    r"""Sort a sequence of numeric values with NaN.

    This function is an extension of the built-in ``sorted`` function.
    It sees NaN values as equivalent to -infinity when the values are
    sorted.

    Args:
        iterable: The numeric values to sort.
        reverse: If set to ``True``, then the list elements are sorted
            as if each comparison were reversed.

    Returns:
        The sorted list.

    Example usage:

    ```pycon

    >>> from grizz.utils.nan import sortnan
    >>> x = [4, float("nan"), 2, 1.2, 7.9, -2]
    >>> sorted(x)
    [4, nan, -2, 1.2, 2, 7.9]
    >>> sortnan(x)
    [nan, -2, 1.2, 2, 4, 7.9]
    >>> sortnan(x, reverse=True)
    [7.9, 4, 2, 1.2, -2, nan]

    ```
    """
    return sorted(iterable, key=lambda x: LowNaN() if math.isnan(x) else x, reverse=reverse)


class LowNaN(float):
    r"""Implement a NaN representation that is always lower than other
    numbers.

    This class is designed to be used to compare numbers with NaN values
    and should not be used in other cases.

    https://docs.python.org/3/library/functions.html#sorted
    """

    def __ge__(self, other: Any) -> bool:
        if not isinstance(other, (float, int)):
            msg = (
                f"'>=' not supported between instances of 'float' and '{type(other).__qualname__}'"
            )
            raise TypeError(msg)
        return False

    def __gt__(self, other: Any) -> bool:
        if not isinstance(other, (float, int)):
            msg = f"'>' not supported between instances of 'float' and '{type(other).__qualname__}'"
            raise TypeError(msg)
        return False

    def __le__(self, other: Any) -> bool:
        if not isinstance(other, (float, int)):
            msg = (
                f"'<=' not supported between instances of 'float' and '{type(other).__qualname__}'"
            )
            raise TypeError(msg)
        return True

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, (float, int)):
            msg = f"'<' not supported between instances of 'float' and '{type(other).__qualname__}'"
            raise TypeError(msg)
        return True



r"""Contain DataFrame columns utility functions."""

from __future__ import annotations

__all__ = [
    "check_column_exist_policy",
    "check_column_missing_policy",
    "check_existing_column",
    "check_existing_columns",
    "check_missing_column",
    "check_missing_columns",
    "find_common_columns",
    "find_missing_columns",
]

import warnings
from typing import TYPE_CHECKING

import polars as pl

from grizz.exceptions import (
    ColumnExistsError,
    ColumnExistsWarning,
    ColumnNotFoundError,
    ColumnNotFoundWarning,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


def check_column_exist_policy(exist_policy: str) -> None:
    r"""Check the policy on how to handle existing columns.

    Args:
        exist_policy: The policy on how to handle existing columns.

    Raises:
        ValueError: if ``exist_policy`` is not ``'ignore'``,
            ``'warn'``, or ``'raise'``.

    Example usage:

    ```pycon

    >>> from grizz.utils.column import check_column_exist_policy
    >>> check_column_exist_policy("ignore")

    ```
    """
    if exist_policy not in {"ignore", "warn", "raise"}:
        msg = (
            f"Incorrect 'exist_policy': {exist_policy}. The valid values are: "
            f"'ignore', 'raise', 'warn'"
        )
        raise ValueError(msg)


def check_column_missing_policy(missing_policy: str) -> None:
    r"""Check the policy on how to handle missing columns.

    Args:
        missing_policy: The policy on how to handle missing columns.

    Raises:
        ValueError: if ``missing_policy`` is not ``'ignore'``,
            ``'warn'``, or ``'raise'``.

    Example usage:

    ```pycon

    >>> from grizz.utils.column import check_column_missing_policy
    >>> check_column_missing_policy("ignore")

    ```
    """
    if missing_policy not in {"ignore", "warn", "raise"}:
        msg = (
            f"Incorrect 'missing_policy': {missing_policy}. The valid values are: "
            f"'ignore', 'raise', 'warn'"
        )
        raise ValueError(msg)


def check_existing_column(
    frame_or_cols: pl.DataFrame | Sequence, column: str, exist_policy: str = "raise"
) -> None:
    r"""Check if a column already exists.

    Args:
        frame_or_cols: The DataFrame or its columns.
        column: The column to check.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised the column already exists. If ``'warn'``,
            a warning is raised if the column already exists and the
            existing column is overwritten. If ``'ignore'``,
            the existing column is overwritten and no warning message
            appears.

    Raises:
        ColumnExistsError: if at least one column already exists and
            ``exist_policy='raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.column import check_existing_column
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> check_existing_column(frame, "col1", exist_policy="warn")

    ```
    """
    check_column_exist_policy(exist_policy)
    exist = column in frame_or_cols
    if not exist:
        return
    msg = f"column {column!r} already exists in the DataFrame"
    if exist_policy == "raise":
        raise ColumnExistsError(msg)
    if exist_policy == "warn":
        warnings.warn(msg + " and will be overwritten", ColumnExistsWarning, stacklevel=2)


def check_existing_columns(
    frame_or_cols: pl.DataFrame | Sequence, columns: Sequence, exist_policy: str = "raise"
) -> None:
    r"""Check if some columns already exist.

    Args:
        frame_or_cols: The DataFrame or its columns.
        columns: The columns to check.
        exist_policy: The policy on how to handle existing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column already exist.
            If ``'warn'``, a warning is raised if at least one column
            already exist and the existing columns are overwritten.
            If ``'ignore'``, the existing columns are overwritten and
            no warning message appears.

    Raises:
        ColumnExistsError: if at least one column already exists and
            ``exist_policy='raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.column import check_existing_columns
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> check_existing_columns(frame, ["col1", "col5"], exist_policy="warn")

    ```
    """
    check_column_exist_policy(exist_policy)
    existing_cols = find_common_columns(frame_or_cols=frame_or_cols, columns=columns)
    if not existing_cols:
        return
    m = "column already exists" if len(existing_cols) == 1 else "columns already exist"
    if exist_policy == "raise":
        msg = f"{len(existing_cols):,} {m} in the DataFrame: {existing_cols}"
        raise ColumnExistsError(msg)
    if exist_policy == "warn":
        msg = (
            f"{len(existing_cols):,} {m} in the DataFrame "
            f"and will be overwritten: {existing_cols}"
        )
        warnings.warn(msg, ColumnExistsWarning, stacklevel=2)


def check_missing_column(
    frame_or_cols: pl.DataFrame | Sequence, column: str, missing_policy: str = "raise"
) -> None:
    r"""Check if a column is missing.

    Args:
        frame_or_cols: The DataFrame or its columns.
        column: The column to check.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if the column is missing. If ``'warn'``,
            a warning is raised if the column is missing and the
            missing columns are ignored. If ``'ignore'``, the missing
            column is ignored and no warning message appears.

    Raises:
        ColumnNotFoundError: if the column is missing and
            ``missing_policy='raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.column import check_missing_column
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> check_missing_column(frame, "col1", missing_policy="warn")

    ```
    """
    check_column_missing_policy(missing_policy)
    exist = column in frame_or_cols
    if exist:
        return
    msg = f"column {column!r} is missing in the DataFrame"
    if missing_policy == "raise":
        raise ColumnNotFoundError(msg)
    if missing_policy == "warn":
        warnings.warn(msg + " and will be ignored", ColumnNotFoundWarning, stacklevel=2)


def check_missing_columns(
    frame_or_cols: pl.DataFrame | Sequence, columns: Sequence, missing_policy: str = "raise"
) -> None:
    r"""Check if some columns are missing.

    Args:
        frame_or_cols: The DataFrame or its columns.
        columns: The columns to check.
        missing_policy: The policy on how to handle missing columns.
            The following options are available: ``'ignore'``,
            ``'warn'``, and ``'raise'``. If ``'raise'``, an exception
            is raised if at least one column is missing.
            If ``'warn'``, a warning is raised if at least one column
            is missing and the missing columns are ignored.
            If ``'ignore'``, the missing columns are ignored and
            no warning message appears.

    Raises:
        ColumnNotFoundError: if at least one column is missing and
            ``missing_policy='raise'``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.column import check_missing_columns
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...         "col4": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> check_missing_columns(frame, ["col1", "col5"], missing_policy="warn")

    ```
    """
    check_column_missing_policy(missing_policy)
    missing_cols = find_missing_columns(frame_or_cols=frame_or_cols, columns=columns)
    if not missing_cols:
        return
    m = "column is" if len(missing_cols) == 1 else "columns are"
    if missing_policy == "raise":
        msg = f"{len(missing_cols):,} {m} missing in the DataFrame: {missing_cols}"
        raise ColumnNotFoundError(msg)
    if missing_policy == "warn":
        msg = (
            f"{len(missing_cols):,} {m} missing in the DataFrame and "
            f"will be ignored: {missing_cols}"
        )
        warnings.warn(msg, ColumnNotFoundWarning, stacklevel=2)


def find_common_columns(
    frame_or_cols: pl.DataFrame | Sequence, columns: Sequence[str]
) -> tuple[str, ...]:
    r"""Find the common columns that are both in the DataFrame and the
    given columns.

    Args:
        frame_or_cols: The DataFrame or its columns.
        columns: The columns to check.

    Returns:
        The columns i.e. the columns that are both in
            ``columns`` and ``frame_or_cols``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.column import find_common_columns
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> cols = find_common_columns(frame, columns=["col1", "col2", "col3", "col4"])
    >>> cols
    ('col1', 'col2', 'col3')

    ```
    """
    cols = set(frame_or_cols.columns if isinstance(frame_or_cols, pl.DataFrame) else frame_or_cols)
    columns = set(columns)
    return tuple(sorted(columns.intersection(cols)))


def find_missing_columns(
    frame_or_cols: pl.DataFrame | Sequence, columns: Sequence[str]
) -> tuple[str, ...]:
    r"""Find the columns that are in the given columns but not in the
    DataFrame.

    Args:
        frame_or_cols: The DataFrame or its columns.
        columns: The columns to check.

    Returns:
        The list of missing columns i.e. the columns that are in
            ``columns`` but not in ``frame_or_cols``.

    Example usage:

    ```pycon

    >>> import polars as pl
    >>> from grizz.utils.column import find_missing_columns
    >>> frame = pl.DataFrame(
    ...     {
    ...         "col1": [1, 2, 3, 4, 5],
    ...         "col2": ["1", "2", "3", "4", "5"],
    ...         "col3": ["a ", " b", "  c  ", "d", "e"],
    ...     }
    ... )
    >>> cols = find_missing_columns(frame, columns=["col1", "col2", "col3", "col4"])
    >>> cols
    ('col4',)

    ```
    """
    cols = set(frame_or_cols.columns if isinstance(frame_or_cols, pl.DataFrame) else frame_or_cols)
    columns = set(columns)
    return tuple(sorted(columns.difference(cols).intersection(columns)))
